#include "links.h"
#include "strsubs.h"

LinkT *SymbolList = 0;

int total_mem_alloc = 0;
int total_mem_freed = 0;

typedef struct mem_alloc_link_s {
    struct mem_alloc_link_s *next;
    void *ptr;
    int size;
} mem_alloc_link_t;

static mem_alloc_link_t *mem_alloc_list = 0;

LinkT *FindToken(char *str)
{
    LinkT *link = 0;
    mem_alloc_link_t *mlink = mem_alloc_list;

    if (str == 0) return link;

    // Locate the string point in the allocated memory list
    while (mlink)
    {
	if (mlink->ptr == str)
	{
	    // The next entry should be the token link
	    mlink = mlink->next;
	    if (mlink)
	    {
	        link = (LinkT *)(mlink->ptr);
	        if (link && link->str != str) link = 0;
	    }
	    break;
	}
	mlink = mlink->next;
    }
    return link;
}

static void mem_alloc_list_add(void *ptr, int size)
{
    mem_alloc_link_t *link = malloc(sizeof(mem_alloc_link_t));
    if (link == 0)
    {
	printf("mem_alloc_list_add: Out of memory\n");
	CspinExit(1);
    }
    link->next = mem_alloc_list;
    link->ptr = ptr;
    link->size = size;
    mem_alloc_list = link;
}

static void mem_alloc_list_remove(void *ptr)
{
    mem_alloc_link_t *prev = 0;
    mem_alloc_link_t *link = mem_alloc_list;

    // Search list for ptr
    while (link)
    {
	if (link->ptr == ptr)
	{
	    if (prev)
	    {
		prev->next = link->next;
	    }
	    else
	    {
		mem_alloc_list = link->next;
	    }
	    total_mem_freed += link->size;
	    free(link);
	    break;
	}
	prev = link;
	link = link->next;
    }
}

void mem_free(void *ptr)
{
    mem_alloc_list_remove(ptr);
    free(ptr);
}

void *mem_alloc(int size)
{
    void *ptr;
    if (size < 0) size = 1;
    ptr = malloc(size);
    if (ptr == 0)
    {
	printf("mem_alloc: Out of memory\n");
	printf("total_mem_alloc = %d\n", total_mem_alloc);
	CspinExit(1);
    }
    mem_alloc_list_add(ptr, size);
    total_mem_alloc += size;
    return ptr;
}

void *mem_alloc_zero(int size)
{
    void *ptr;
    if (size < 0) size = 1;
    ptr = mem_alloc(size);
    memset(ptr, 0, size);
    return ptr;
}

char *str_alloc(int size)
{
    char *ptr = mem_alloc(size);
    *ptr = 0;
    return ptr;
}

LinkT *LinkCreate(char *str, int size)
{
    LinkT *link = mem_alloc_zero(size);
    link->next = 0;
    link->data = 0;
    link->id = 0;
    if (str == 0)
        link->str = 0;
    else
    {
	link->str = mem_alloc(strlen(str) + 1);
	strcpy(link->str, str);
    }
    return link;
}

void LinkDestroy(LinkT *link)
{
    if (link == 0) return;
    if (link->str != 0) mem_free(link->str);
    mem_free(link);
}

LinkT *ListDestroy(LinkT *list)
{
    LinkT *link;
    while (list)
    {
	link = list;
	list = list->next;
        if (link->str != 0) mem_free(link->str);
        mem_free(link);
    }
    return 0;
}

LinkT *ListTail(LinkT *list)
{
    if (list == 0) return 0;
    while (list->next) list = list->next;
    return list;
}

LinkT *ListAppend(LinkT *list, LinkT *link)
{
    LinkT *tail;
    if (list == 0) list = link;
    else
    {
        tail = ListTail(list);
        tail->next = link;
    }
    return list;
}

LinkT *ListInsertAfter(LinkT *list, LinkT *link)
{
    LinkT *next;
    if (list == 0) return link;
    if (link == 0) return list;
    next = list->next;
    list->next = link;
    link = ListTail(link);
    link->next = next;
    return list;
}

LinkT *ListFindPrev(LinkT *list, LinkT *link)
{
    if (list == 0) return 0;
    if (link == 0) return 0;
    if (list == link) return 0;
    while (list->next && list->next != link) list = list->next;
    return list;
}

LinkT *ListRemoveLink(LinkT *list, LinkT *link)
{
    LinkT *prev;
    if (list == 0) return list;
    if (link == 0) return list;
    if (list == link)
    {
	list = link->next;
	link->next = 0;
	return list;
    }
    prev = ListFindPrev(list, link);
    if (prev == 0) return list;
    prev->next = link->next;
    link->next = 0;
    return list;
}

LinkT *ListAppendLinkCreate(LinkT *list, char *str, int size)
{
    LinkT *link = LinkCreate(str, size);
    LinkT *tail;
    if (list == 0) return link;
    tail = ListTail(list);
    tail->next = link;
    return list;
}

LinkT *ListDuplicate(LinkT *list)
{
    LinkT *link;
    LinkT *tail = 0;
    LinkT *dupe = 0;

    while (list)
    {
        link = LinkCreate(list->str, sizeof(LinkT));
        if (tail) tail->next = link;
        else      dupe = link;
        tail = link;
        list = list->next;
    }

    return dupe;
}

void ListPrint(LinkT *list)
{
    int i = 0;
    while (list)
    {
	if (list->str)
	    printf("%d: %s\n", i++, list->str);
	else
	    printf("%d: NULL\n", i++);
	list = list->next;
    }
}

LinkT *ListFindLinkString(LinkT *list, char *str)
{
    if (str == 0) return 0;

    while (list)
    {
	if (strcmp(list->str, str) == 0) break;
	list = list->next;
    }

    return list;
}

LinkT *ListFindLinkStringLast(LinkT *list, char *str)
{
    LinkT *link = 0;

    if (str == 0) return link;

    while (list)
    {
	if (strcmp(list->str, str) == 0) link = list;
	list = list->next;
    }

    return link;
}

LinkT *ListRemoveLinkString(LinkT *list, char *str)
{
    LinkT *link = ListFindLinkString(list, str);
    return ListRemoveLink(list, link);
}

LinkT *ListFindLinkStringNoCase(LinkT *list, char *str)
{
    if (str == 0) return list;
    while (list)
    {
	if (strcmpnocase(list->str, str) == 0) break;
	list = list->next;
    }
    return list;
}

char *AddToSymbolList(char *str)
{
    int val;
    int len;
    LinkT *link;
    char str1[31];

    strncpy(str1, str, 30);
    str1[30] = 0;
    link = ListFindLinkStringNoCase(SymbolList, str1);

    if (link)
    {
	//printf("AddToSymbolList: %s, found %s\n", str, link->str);
	len = strlen(str1);
	if (len > 28) len = 28;
	str1[len++] = '_';
	str1[len+1] = 0;
	val = '0' - 1;
    }

    // Generate a unique name if already on the list
    while (link)
    {
	if (val == 'z')
	{
	    printf("Could not generate a unique name for %s\n", str);
	    CspinExit(1);
	}
	val++;
	if (val == '9' + 1) val = 'a';
	str1[len] = val;
        link = ListFindLinkStringNoCase(SymbolList, str1);
	//if (link)
	//printf("AddToSymbolList: %s, found %s\n", str2, link->str);
    }

    //if (strcmp(str, str1))
        //printf("AddToSymbolList: changed %s to %s\n", str, str1);

    //printf("AddToSymbolList: Added \"%s\" as \"%s\"\n", str, str1);
    //printf("AddToSymbolList: %s\n", str2);
    link = LinkCreate(str1, sizeof(LinkT));
    SymbolList = ListAppend(SymbolList, link);

    return link->str;
}

void RemoveFromSymbolList(char *str)
{
    LinkT *link;

    if (str == 0)
    {
	printf("RemoveFromSymbolList: NULL pointer\n");
	return;
    }

    link = ListFindLinkString(SymbolList, str);

    if (link == 0)
    {
	printf("RemoveFromSymbolList: %s is not in symbol list\n", str);
	return;
    }

    SymbolList = ListRemoveLink(SymbolList, link);
    //printf("RemoveFromSymbolList: Removed %s from symbol list\n", str);
}
